# -*- coding: utf-8 -*-
"""
validator handlers base module.
"""

import inspect

from sqlalchemy.orm.attributes import InstrumentedAttribute

import pyrin.utils.misc as misc_utils

from pyrin.core.globals import _, LIST_TYPES
from pyrin.database.model.base import BaseEntity
from pyrin.validator.exceptions import ValidationError
from pyrin.validator.interface import AbstractValidatorBase
from pyrin.validator.handlers.exceptions import ValidatorFieldIsRequiredError, \
    ValueCouldNotBeNoneError, InvalidValueTypeError, InvalidValidatorDomainError, \
    InvalidAcceptedTypeError, InvalidValidationExceptionTypeError, ValueIsNotListError, \
    ValidatorFixerMustBeCallable, ValueCouldNotBeAnEmptyListError


class ValidatorBase(AbstractValidatorBase):
    """
    validator base class.

    all application validators must be subclassed from this.
    """

    invalid_type_error = InvalidValueTypeError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'is not an instance of {type}.')
    none_value_error = ValueCouldNotBeNoneError
    none_value_message = _('The provided value for [{param_name}] could not be None.')
    not_list_error = ValueIsNotListError
    not_list_message = _('The provided value for [{param_name}] must be a list of items.')
    empty_list_error = ValueCouldNotBeAnEmptyListError
    empty_list_message = _('The provided value for [{param_name}] could not be an empty list.')

    # specifies that the value could be None.
    default_nullable = None

    # localized name of parameter that this validator will validate.
    default_localized_name = None

    # specifies that value must be a list of items.
    default_is_list = False

    # specifies that list validator should also accept single, non list values.
    # it only has effect if 'default_is_list=True' is set.
    default_allow_single = False

    # specifies that list items could also be None.
    # it only has effect if 'default_is_list=True' is set.
    default_null_items = False

    # specifies that list values could be an empty list.
    # it only has effect if 'default_is_list=True' is set.
    default_allow_empty_list = False

    # a callable that accepts a single value and returns the fixed or the same value.
    default_fixer = None

    def __init__(self, domain, field, **options):
        """
        initializes an instance of ValidatorBase.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param InstrumentedAttribute | str field: validator field name. it could be a
                                                  string or a column. each validator will
                                                  be registered with its field name in
                                                  corresponding domain. to enable automatic
                                                  validations, the provided field name must
                                                  be the exact name of the parameter which
                                                  this validator will validate. if you pass
                                                  a column attribute, some constraints
                                                  such as `nullable`, `min_length`, `max_length`,
                                                  `min_value`, `max_value`, `allow_blank`,
                                                  `allow_whitespace`, `check_in` and
                                                  `check_not_in` could be extracted
                                                  automatically from that column if not provided
                                                  in inputs.

        :keyword type | tuple[type] accepted_type: accepted type for value.
                                                   no type checking will be
                                                   done if not provided.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValidatorFieldIsRequiredError: validator field is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises InvalidAcceptedTypeError: invalid accepted type error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        is_column = isinstance(field, InstrumentedAttribute)
        is_str = isinstance(field, str)
        name = field
        if is_column:
            name = field.key

        if (not is_column and not is_str) or (is_str and (field.isspace() or field == '')):
            raise ValidatorFieldIsRequiredError('Validator field must be provided for '
                                                'validator [{instance}]. it could be '
                                                'a string or a [{base}] type.'
                                                .format(instance=self,
                                                        base=InstrumentedAttribute))

        if (not inspect.isclass(domain) or not issubclass(domain, BaseEntity)) and \
                (not isinstance(domain, str) or domain.isspace()):
            raise InvalidValidatorDomainError('The provided domain for validator [{instance}] '
                                              'with name [{name}] is not a subclass of '
                                              '[{entity}] or a string value.'
                                              .format(instance=self, name=name,
                                                      entity=BaseEntity))

        nullable = options.get('nullable')
        if nullable is None:
            if self.default_nullable is not None:
                nullable = self.default_nullable
            elif is_column:
                nullable = field.nullable
            else:
                nullable = True

        is_list = options.get('is_list')
        if is_list is None:
            if self.default_is_list is not None:
                is_list = self.default_is_list
            else:
                is_list = False

        null_items = options.get('null_items')
        if null_items is None:
            if self.default_null_items is not None:
                null_items = self.default_null_items
            elif is_column:
                null_items = field.nullable
            else:
                null_items = False

        allow_single = options.get('allow_single')
        if allow_single is None:
            if self.default_allow_single is not None:
                allow_single = self.default_allow_single
            else:
                allow_single = False

        allow_empty_list = options.get('allow_empty_list')
        if allow_empty_list is None:
            if self.default_allow_empty_list is not None:
                allow_empty_list = self.default_allow_empty_list
            else:
                allow_empty_list = False

        accepted_type = options.get('accepted_type')
        if accepted_type is not None and not isinstance(accepted_type, (type, tuple)):
            raise InvalidAcceptedTypeError('The provided accepted type '
                                           '[{accepted_type}] must be a type '
                                           'or tuple of types.'
                                           .format(accepted_type=accepted_type))

        if isinstance(accepted_type, tuple):
            if len(accepted_type) <= 0:
                raise InvalidAcceptedTypeError('The provided accepted type tuple '
                                               'should have at least one item in it.'
                                               'if no type checking should be done, '
                                               'do not provide the accepted type '
                                               'keyword argument.')
            for item in accepted_type:
                if not isinstance(item, type):
                    raise InvalidAcceptedTypeError('The provided accepted type '
                                                   '[{accepted_type}] must be a type'
                                                   .format(accepted_type=accepted_type))

        if self.default_fixer is not None and not callable(self.default_fixer):
            raise ValidatorFixerMustBeCallable('Validator [{name}] fixer must be a '
                                               'callable with single argument.'
                                               .format(name=name))

        localized_name = options.get('localized_name') or self.default_localized_name
        if localized_name in (None, '') or localized_name.isspace():
            localized_name = _(name)

        self._validate_exception_type(self.invalid_type_error)
        self._validate_exception_type(self.none_value_error)
        self._validate_exception_type(self.not_list_error)

        super().__init__()

        self._field = None
        if is_column:
            self._field = field

        self._set_name(name)
        self._domain = domain
        self._nullable = nullable
        self._accepted_type = accepted_type
        self._localized_name = localized_name
        self._is_list = is_list
        self._null_items = null_items
        self._allow_single = allow_single
        self._allow_empty_list = allow_empty_list

    def validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        it returns the same or fixed value.

        :param object | list[object] value: value to be validated.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :keyword bool nullable: determines that provided value could be None.
                                this value has precedence over `nullable`
                                instance attribute if provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               this value has precedence over `is_list`
                               instance attribute if provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  this value has precedence over `null_items`
                                  instance attribute if provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValueIsNotListError: value is not list error.
        :raises InvalidValueTypeError: invalid value type error.
        :raises ValueCouldNotBeNoneError: value could not be none error.
        :raises ValidationError: validation error.

        :returns: object | list[object]
        """

        nullable = options.pop('nullable', None)
        if nullable is None:
            nullable = self.nullable

        is_list = options.pop('is_list', None)
        if is_list is None:
            is_list = self.is_list

        null_items = options.pop('null_items', None)
        if null_items is None:
            null_items = self.null_items

        allow_single = options.pop('allow_single', None)
        if allow_single is None:
            allow_single = self.allow_single

        allow_empty_list = options.pop('allow_empty_list', None)
        if allow_empty_list is None:
            allow_empty_list = self.allow_empty_list

        if value is None:
            self._validate_nullable(value, nullable, **options)
            return value

        is_really_list = isinstance(value, list)
        if is_list is True and is_really_list is True \
                and len(value) <= 0 and allow_empty_list is False:
            raise self.empty_list_error(
                self.empty_list_message.format(param_name=self.localized_name))

        if is_list is True and allow_single is False and is_really_list is False:
            raise self.not_list_error(
                self.not_list_message.format(param_name=self.localized_name))

        if is_list is False or (is_list is True and is_really_list is False):
            return self._perform_validation(value, nullable, **options)
        else:
            temp_values = list(value)
            for index, item in enumerate(temp_values):
                value[index] = self._perform_validation(item, null_items, **options)

            return value

    def _perform_validation(self, value, nullable, **options):
        """
        performs validation on given value.

        it raises an error if validation fails.
        it returns the same or fixed value.

        :param object value: value to be validated.
        :param bool nullable: determines that provided value could be None.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :raises InvalidValueTypeError: invalid value type error.
        :raises ValueCouldNotBeNoneError: value could not be none error.
        :raises ValidationError: validation error.

        :returns: object
        """

        if value is None:
            self._validate_nullable(value, nullable, **options)
            return value

        value = self._validate_type(value)
        self._validate(value, **options)
        return value

    def _validate_nullable(self, value, nullable, **options):
        """
        validates value for nullability.

        it raises an error if validation fails.

        :param object value: value to be validated.
        :param bool nullable: determines that provided value could be None.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :raises ValueCouldNotBeNoneError: value could not be none error.
        """

        for_update = options.get('for_update', False)
        has_default = False
        if self.field is not None:
            if for_update is True:
                has_default = self.field.onupdate is not None or \
                              self.field.server_onupdate is not None
            else:
                has_default = self.field.default is not None or \
                              self.field.server_default is not None

        if nullable is False and value is None and has_default is False:
            raise self.none_value_error(self.none_value_message.format(
                param_name=self.localized_name))

    def _validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        the raised error must be an instance of ValidationError.
        this method must be implemented in subclasses.
        each overridden method must call `super()._validate()`
        preferably at the beginning.
        if no extra validation is needed, it could be left unimplemented.

        :param object value: value to be validated.

        :raises ValidationError: validation error.
        """
        pass

    def _validate_type(self, value):
        """
        validates the type of given value.

        if no accepted type is set for this validator, this method does nothing.
        if the value type is not the same as accepted type but this validator
        has a fixer set for it, it fixes the type using that fixer.
        it returns the same input or fixed one.

        :param object value: value to be validated.

        :raises InvalidValueTypeError: invalid value type error.

        :returns: object
        """

        if self.accepted_type is None:
            if self.default_fixer is not None:
                return self._fix_value(value)

            return value

        is_accepted = isinstance(value, self.accepted_type)
        if not is_accepted:
            if self.default_fixer is not None:
                value = self._fix_value(value)
                is_accepted = isinstance(value, self.accepted_type)

            if not is_accepted:
                preview_type = misc_utils.make_iterable(self.accepted_type, list)

                raise self.invalid_type_error(self.invalid_type_message.format(
                    param_name=self.localized_name, type=preview_type))

        return value

    def _validate_exception_type(self, exception):
        """
        asserts that given exception type is subclassed from ValidationError.

        :param type[CoreException] exception: exception to validate its type.

        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        if not issubclass(exception, ValidationError):
            raise InvalidValidationExceptionTypeError('The specified validation exception '
                                                      '[{exception}] is not a subclass of '
                                                      '[{base}].'
                                                      .format(exception=exception,
                                                              base=ValidationError))

    def _get_representation(self, value):
        """
        gets the string representable version of input value.

        :param object value: value to get its string representation.

        :returns: string representable value.
        """

        if self.accepted_type is None or not isinstance(value, self.accepted_type):
            return value

        return self._get_safe_representation(value)

    def _get_safe_representation(self, value):
        """
        gets the string representable version of input value.

        this method is intended to be overridden in subclasses for complex types.
        the type of value is guaranteed to be what subclasses are expecting.

        :param object value: value to get its string representation.

        :returns: string representable value.
        """

        return value

    def _get_list_representation(self, value):
        """
        gets the string representable version of input list.

        :param list value: value to get its string representation.

        :returns: string representable value.
        :rtype: list
        """

        if not isinstance(value, LIST_TYPES):
            return self._get_representation(value)

        result = []
        for item in value:
            result.append(self._get_representation(item))

        return result

    @classmethod
    def _fix_value(cls, value):
        """
        fixes the given value using this validator's fixer.

        :param object value: value to be fixed.

        :returns: fixed value.
        """

        return cls._get_value(cls.default_fixer, value)

    @classmethod
    def _get_value(cls, value, *args):
        """
        gets the value of given input.

        it is a helper method which could call the input if
        it is a callable or get the same input.

        :param object | callable value: value to get its original value.
                                        it could be an object or a callable.

        :param object args: the inputs that given callable requires.

        :returns: result of callable input or the same input.
        """

        if callable(value):
            return value(*args)

        return value

    @property
    def name(self):
        """
        gets the name of this validator.

        :rtype: str
        """

        return self.get_name()

    @property
    def domain(self):
        """
        gets the domain of this validator.

        domain is the type of a BaseEntity subclass that
        this validator validates a value of it.
        if a validator is not specific to an entity, then
        domain could be a unique string name.

        :rtype: type[BaseEntity] | str
        """

        return self._domain

    @property
    def nullable(self):
        """
        gets a value indicating that null values must be accepted as valid.

        :rtype: bool
        """

        return self._nullable

    @property
    def accepted_type(self):
        """
        gets the accepted type for this validator.

        returns None if no type checking is defined for this validator.

        :rtype: type | tuple[type]
        """

        return self._accepted_type

    @property
    def localized_name(self):
        """
        gets the localized name of this validator.

        :rtype: str
        """

        return self._localized_name

    @property
    def is_list(self):
        """
        gets a value indicating that values must be a list of items.

        :rtype: bool
        """

        return self._is_list

    @property
    def null_items(self):
        """
        gets a value indicating that list items could be None.

        :rtype: bool
        """

        return self._null_items

    @property
    def allow_single(self):
        """
        gets a value indicating that list validator should also accept single, non list values.

        :rtype: bool
        """

        return self._allow_single

    @property
    def allow_empty_list(self):
        """
        gets a value indicating that list validator should also accept empty lists.

        :rtype: bool
        """

        return self._allow_empty_list

    @property
    def field(self):
        """
        gets the field that this validator will validate.

        it could be None if the provided field for this validator was a string.

        :rtype: InstrumentedAttribute
        """

        return self._field



# -*- coding: utf-8 -*-
"""
authorization component module.
"""

from pyrin.application.decorators import component
from pyrin.security.authorization import AuthorizationPackage
from pyrin.security.authorization.manager import AuthorizationManager
from pyrin.application.structs import Component


@component(AuthorizationPackage.COMPONENT_NAME)
class AuthorizationComponent(Component, AuthorizationManager):
    """
    authorization component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
session component module.
"""

from pyrin.application.decorators import component
from pyrin.security.session import SessionPackage
from pyrin.security.session.manager import SessionManager
from pyrin.application.structs import Component


@component(SessionPackage.COMPONENT_NAME)
class SessionComponent(Component, SessionManager):
    """
    session component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm sql schema columns module.
"""

from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy import BigInteger, Integer, Sequence, ForeignKey, String, Unicode

import pyrin.utils.misc as misc_utils
import pyrin.utils.unique_id as uuid_utils

from pyrin.database.orm.types.custom import GUID
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.orm.sql.schema.exceptions import SequencePKColumnTypeIsInvalidError, \
    AutoPKColumnTypeIsInvalidError, InvalidFKColumnReferenceTypeError, \
    StringColumnTypeIsInvalidError


class StringColumn(CoreColumn):
    """
    string column class.

    this is a helper class to define columns that their type
    is an instance or subclass of sqlalchemy `String`.
    """

    DEFAULT_TYPE = Unicode

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of StringColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.
                                 the type must be an instance or subclass of sqlalchemy
                                 `String` type. defaults to `Unicode` if not provided.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.

        :keyword callable | object onupdate: a scalar, python callable, or
                                             `ClauseElement` representing a default
                                             value to be applied to the column within update
                                             statements, which will be invoked upon update
                                             if this column is not present in the set
                                             clause of the update.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword object server_default: a `FetchedValue` instance, str, unicode
                                        or `text` construct representing the ddl
                                        default value for the column.

        :keyword FetchedValue server_onupdate: a `FetchedValue` instance representing a
                                               database-side default generation function,
                                               such as a trigger. this indicates to sqlalchemy
                                               that a newly generated value will be available
                                               after updates. this construct does not actually
                                               implement any kind of generation function within
                                               the database, which instead must be specified
                                               separately.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to True if not provided.

        :keyword int min_length: minimum length of value for this column.
                                 defaults to None if not provided.

        :keyword int max_length: maximum length of value for this column.
                                 if provided, and the type of this column is a
                                 class, it will be instantiated with this length.
                                 defaults to None if not provided.

        :keyword bool allow_blank: specifies that this column could have blank string
                                   value. defaults to False if not provided.

        :keyword bool allow_whitespace: specifies that this column could have whitespace
                                        string value. defaults to False if not provided.

        :keyword object | callable min_value: minimum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword object | callable max_value: maximum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword list | callable check_in: list of valid values for this column.
                                           it could also be a callable without any inputs.
                                           if a non-callable is provided and the column is
                                           not a primary key and also column name is provided,
                                           it will result in check constraint generation on
                                           database. otherwise it will be ignored and could
                                           be used in validators.
                                           defaults to None if not provided.

        :keyword list | callable check_not_in: list of invalid values for this column.
                                               it could also be a callable without any inputs.
                                               if a non-callable is provided and the column is
                                               not a primary key and also column name is provided,
                                               it will result in check constraint generation on
                                               database. otherwise it will be ignored and could
                                               be used in validators.
                                               defaults to None if not provided.

        :note check_in, check_not_in: only one of these options could be provided.
                                      otherwise it raises an error.

        :raises StringColumnTypeIsInvalidError: string column type is invalid error.
        """

        self.min_length = kwargs.pop('min_length', None)
        self.max_length = kwargs.pop('max_length', None)
        self.allow_blank = kwargs.pop('allow_blank', False)
        self.allow_whitespace = kwargs.pop('allow_whitespace', False)

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)
        if type_ is None:
            type_ = self.DEFAULT_TYPE

        if not misc_utils.is_subclass_or_instance(type_, String):
            raise StringColumnTypeIsInvalidError('The string column type must be '
                                                 'an instance or subclass of [{string}].'
                                                 .format(string=String))

        if isinstance(type_, type) and self.max_length is not None:
            type_ = type_(length=self.max_length)

        kwargs.update(name=name, type_=type_)

        super().__init__(*args, **kwargs)


class PKColumn(CoreColumn):
    """
    pk column class.

    this is a helper class for defining pk columns.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of PKColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.
                             defaults to True if not provided.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword object server_default: a `FetchedValue` instance, str, unicode
                                        or `text` construct representing the ddl
                                        default value for the column.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to False if not provided.
        """

        kwargs.update(nullable=False, primary_key=True)
        kwargs.setdefault('allow_write', False)
        kwargs.setdefault('index', True)
        kwargs.pop('onupdate', None)
        kwargs.pop('server_onupdate', None)
        kwargs.pop('unique', None)

        super().__init__(*args, **kwargs)


class AutoPKColumn(PKColumn):
    """
    auto pk column class.

    this is a helper class for defining pk columns with auto increment value.
    this type of pk column's value is not available to python side without commit or flush.
    """

    DEFAULT_TYPE = BigInteger

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of AutoPKColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 it must be an instance or subclass of `Integer` type.
                                 defaults to `DEFAULT_TYPE` if not provided.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.
                             defaults to True if not provided.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to False if not provided.

        :raises AutoPKColumnTypeIsInvalidError: auto pk column type is invalid error.
        """

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)
        if type_ is None:
            type_ = self.DEFAULT_TYPE

        if not misc_utils.is_subclass_or_instance(type_, Integer):
            raise AutoPKColumnTypeIsInvalidError('The auto pk column type must be an '
                                                 'instance or subclass of [{integer}].'
                                                 .format(integer=Integer))

        kwargs.update(name=name, type_=type_, autoincrement=True, min_value=1)
        kwargs.pop('default', None)
        kwargs.pop('server_default', None)

        super().__init__(*args, **kwargs)


class GUIDPKColumn(PKColumn):
    """
    guid pk column class.

    this is a helper class for defining pk columns that their value is a guid.
    this type of pk column's value is available to python side without commit or flush.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of GUIDPKColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.
                             defaults to True if not provided.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to False if not provided.
        """

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)
        kwargs.update(name=name, type_=GUID, autoincrement=False,
                      default=uuid_utils.generate_uuid4)
        kwargs.pop('server_default', None)

        super().__init__(*args, **kwargs)


class SequencePKColumn(PKColumn):
    """
    sequence pk column class.

    this is a helper class for defining pk columns that gain their value from a sequence.
    it differs from columns that set `autoincrement=True`, because the value of sequence
    columns is available to python side without commit or flush. and also a table can have
    multiple sequence columns which is impossible for auto increment columns.
    """

    DEFAULT_CACHE = 100
    DEFAULT_TYPE = BigInteger

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of SequencePKColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 it must be an instance or subclass of `Integer` type.
                                 defaults to `DEFAULT_TYPE` if not provided.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword str sequence: sequence name to be generated for this column.
                               this value is required, but has to be set as
                               keyword to prevent errors.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.
                             defaults to True if not provided.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to False if not provided.

        :keyword int cache: cache size for sequence.
                            defaults to `DEFAULT_CACHE`, if not provided.
                            to disable cache, you can pass it as None or `0`.
                            note that cache is per session, so if you stop the
                            connection to database and start it again, the new
                            session will get its own cache. but it's good for
                            performance to have cache on sequences, gaps are not
                            bad at all.

        :raises SequencePKColumnTypeIsInvalidError: sequence pk column type is invalid error.
        """

        args = list(args)
        name, type_ = self._extract_name_and_type(args, kwargs)
        if type_ is None:
            type_ = self.DEFAULT_TYPE

        if not misc_utils.is_subclass_or_instance(type_, Integer):
            raise SequencePKColumnTypeIsInvalidError('The sequence pk column type must be an '
                                                     'instance or subclass of [{integer}].'
                                                     .format(integer=Integer))

        cache = kwargs.pop('cache', self.DEFAULT_CACHE)
        sequence = kwargs.pop('sequence', None)
        sequence_kwargs = dict()
        if cache is not None and cache > 0:
            sequence_kwargs.update(cache=cache)

        sequence_instance = Sequence(sequence, **sequence_kwargs)
        # this is to prevent sqlalchemy errors.
        if sequence is not None:
            args.append(sequence_instance)

        kwargs.update(name=name, type_=type_,
                      autoincrement=False, min_value=1,
                      default=sequence_instance,
                      server_default=sequence_instance.next_value())

        super().__init__(*args, **kwargs)


class FKColumn(CoreColumn):
    """
    fk column class.

    this is a helper class for defining fk columns.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of FKColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword str | CoreColumn fk: reference column name or instance.
                                      this parameter is required but to
                                      prevent errors it has to be set as keyword.

        :keyword str fk_on_update: optional string. if set, emit ON UPDATE <value> when
                                   issuing DDL for this constraint. typical values include
                                   CASCADE, DELETE and RESTRICT.

        :keyword str fk_on_delete: optional string. if set, emit ON DELETE <value> when
                                   issuing DDL for this constraint. typical values include
                                   CASCADE, DELETE and RESTRICT.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.
                             defaults to True if not provided.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.
                                default to False if not provided.

        :keyword callable | object onupdate: a scalar, python callable, or
                                             `ClauseElement` representing a default
                                             value to be applied to the column within update
                                             statements, which will be invoked upon update
                                             if this column is not present in the set
                                             clause of the update.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword object server_default: a `FetchedValue` instance, str, unicode
                                        or `text` construct representing the ddl
                                        default value for the column.

        :keyword FetchedValue server_onupdate: a `FetchedValue` instance representing a
                                               database-side default generation function,
                                               such as a trigger. this indicates to sqlalchemy
                                               that a newly generated value will be available
                                               after updates. this construct does not actually
                                               implement any kind of generation function within
                                               the database, which instead must be specified
                                               separately.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to True if not provided.

        :keyword object | callable min_value: minimum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword object | callable max_value: maximum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword list | callable check_in: list of valid values for this column.
                                           it could also be a callable without any inputs.
                                           if a non-callable is provided and the column is
                                           not a primary key and also column name is provided,
                                           it will result in check constraint generation on
                                           database. otherwise it will be ignored and could
                                           be used in validators.
                                           defaults to None if not provided.

        :keyword list | callable check_not_in: list of invalid values for this column.
                                               it could also be a callable without any inputs.
                                               if a non-callable is provided and the column is
                                               not a primary key and also column name is provided,
                                               it will result in check constraint generation on
                                               database. otherwise it will be ignored and could
                                               be used in validators.
                                               defaults to None if not provided.

        :note check_in, check_not_in: only one of these options could be provided.
                                      otherwise it raises an error.
        """

        self._fk = kwargs.pop('fk', None)
        self._fk_on_update = kwargs.pop('fk_on_update', None)
        self._fk_on_delete = kwargs.pop('fk_on_delete', None)

        kwargs.setdefault('index', True)
        kwargs.setdefault('nullable', False)

        super().__init__(*args, **kwargs)

    def _get_custom_schema_items(self):
        """
        gets custom schema items for this column.

        it will generate required fk constraint.

        :raises InvalidFKColumnReferenceTypeError: invalid fk column reference type error.

        :rtype: list
        """

        if self._fk is not None and not isinstance(self._fk, (str,
                                                              InstrumentedAttribute,
                                                              CoreColumn)):
            raise InvalidFKColumnReferenceTypeError('Foreign key reference column '
                                                    'must be a string or a column instance.')

        # this is to prevent sqlalchemy errors.
        # because metadata uses uninitialized entities.
        if self._fk is None:
            self._fk = ''

        return [ForeignKey(self._fk,
                           onupdate=self._fk_on_update,
                           ondelete=self._fk_on_delete)]


class HiddenColumn(CoreColumn):
    """
    hidden column class.

    this is a helper class for defining hidden columns.
    hidden columns will not be included in entity to dict conversion
    and also they won't get populated on conversion from dict.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of HiddenColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.

        :keyword callable | object onupdate: a scalar, python callable, or
                                             `ClauseElement` representing a default
                                             value to be applied to the column within update
                                             statements, which will be invoked upon update
                                             if this column is not present in the set
                                             clause of the update.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword object server_default: a `FetchedValue` instance, str, unicode
                                        or `text` construct representing the ddl
                                        default value for the column.

        :keyword FetchedValue server_onupdate: a `FetchedValue` instance representing a
                                               database-side default generation function,
                                               such as a trigger. this indicates to sqlalchemy
                                               that a newly generated value will be available
                                               after updates. this construct does not actually
                                               implement any kind of generation function within
                                               the database, which instead must be specified
                                               separately.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword object | callable min_value: minimum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword object | callable max_value: maximum value that this column could have.
                                              it could also be a callable without any inputs.
                                              if a non-callable is provided and the column is
                                              not a primary key and also column name is provided,
                                              it will result in check constraint generation on
                                              database. otherwise it will be ignored and could
                                              be used in validators.
                                              defaults to None if not provided.

        :keyword list | callable check_in: list of valid values for this column.
                                           it could also be a callable without any inputs.
                                           if a non-callable is provided and the column is
                                           not a primary key and also column name is provided,
                                           it will result in check constraint generation on
                                           database. otherwise it will be ignored and could
                                           be used in validators.
                                           defaults to None if not provided.

        :keyword list | callable check_not_in: list of invalid values for this column.
                                               it could also be a callable without any inputs.
                                               if a non-callable is provided and the column is
                                               not a primary key and also column name is provided,
                                               it will result in check constraint generation on
                                               database. otherwise it will be ignored and could
                                               be used in validators.
                                               defaults to None if not provided.

        :note check_in, check_not_in: only one of these options could be provided.
                                      otherwise it raises an error.
        """

        kwargs.update(allow_read=False)
        kwargs.update(allow_write=False)

        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
model exceptions module.
"""

from pyrin.core.exceptions import CoreException


class DatabaseModelException(CoreException):
    """
    database model exception.
    """
    pass


class ColumnNotExistedError(DatabaseModelException):
    """
    column not existed error.
    """
    pass


class InvalidDeclarativeBaseTypeError(DatabaseModelException):
    """
    invalid declarative base type error.
    """
    pass


class InvalidDepthProvidedError(DatabaseModelException):
    """
    invalid depth provided error.
    """
    pass


class EntitiesAreNotCollectedError(DatabaseModelException):
    """
    entities are not collected error.
    """
    pass


class InvalidModelHookTypeError(DatabaseModelException):
    """
    invalid model hook type error.
    """
    pass



# -*- coding: utf-8 -*-
"""
sentry integrations celery module.
"""

import pyrin.configuration.services as config_services

from pyrin.logging.sentry.decorators import sentry_integration
from pyrin.logging.sentry.integrations.base import SentryIntegrationBase


@sentry_integration()
class CelerySentryIntegration(SentryIntegrationBase):
    """
    celery sentry integration class.
    """

    name = 'celery'

    def _configure(self, **options):
        """
        configures this integration and gets the configured object.

        :rtype: CeleryIntegration
        """

        # we have to import CeleryIntegration here, because sentry imports celery
        # at the module level and raises an error if it isn't installed.
        from sentry_sdk.integrations.celery import CeleryIntegration

        propagate_traces = config_services.get('sentry', self.get_name(), 'propagate_traces')
        return CeleryIntegration(propagate_traces)



# -*- coding: utf-8 -*-
"""
mimetype component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.processor.mimetype import MIMETypePackage
from pyrin.processor.mimetype.manager import MIMETypeManager


@component(MIMETypePackage.COMPONENT_NAME)
class MIMETypeComponent(Component, MIMETypeManager):
    """
    mimetype component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
bcrypt hashing handler module.
"""

import re

import bcrypt

import pyrin.configuration.services as config_services

from pyrin.security.hashing.decorators import hashing
from pyrin.security.hashing.handlers.base import HashingBase
from pyrin.security.hashing.handlers.exceptions import BcryptMaxSizeLimitError


@hashing()
class BcryptHashing(HashingBase):
    """
    bcrypt hashing class.
    """

    # regular expression to validate format of full hashed values.
    # the following format will be matched:
    # `$handler_name$prefix$rounds$salt-text_plus_salt_hash`
    FORMAT_REGEX = re.compile(r'^\$bcrypt\$[^$]+\$[\d]+\$(.+)$')

    # bcrypt hashing has a max size limit that could be hashed, original
    # implementation truncates the input string to satisfy max size limit.
    # since the max size is different on each bcrypt implementation,
    # we consider 50 bytes as a safe max size for all implementations.
    # 50 bytes is equivalent to 50 chars in ascii encoding, not unicode.
    MAX_SIZE = 50

    def __init__(self, **options):
        """
        initializes an instance of BcryptHashing.
        """

        super().__init__(**options)

    def _generate_hash(self, text, **options):
        """
        gets the hash of input text using a random or specified salt.

        :param str text: text to be hashed.

        :keyword bytes salt: salt to be used for hashing.
                             if not provided, a random salt will be generated.

        :keyword str prefix: prefix to be used for hashing.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :raises BcryptMaxSizeLimitError: bcrypt max size limit error.

        :rtype: bytes
        """

        text_bytes = self._digest_inputs(text, **options)

        salt = options.get('salt', None)
        if salt is None:
            salt = self._generate_salt(**options)

        bcrypt_hash = bcrypt.hashpw(text_bytes, salt)
        return self._make_final_hash(bcrypt_hash)

    def _generate_salt(self, **options):
        """
        generates a valid salt for this handler and returns it.

        :keyword str prefix: prefix to be used for hashing.

        :keyword int rounds: rounds to perform for generating hash.
                             if not provided, default value from
                             relevant config will be used.

        :rtype: bytes
        """

        rounds = options.get('rounds', config_services.get('security', 'hashing',
                                                           'bcrypt_rounds'))

        prefix = options.get('prefix', b'2b')

        return bcrypt.gensalt(rounds=rounds, prefix=prefix)

    def _is_match(self, text, hashed_value, **options):
        """
        gets a value indicating that given text's
        hash is identical to given hashed value.

        :param str text: text to be hashed.
        :param bytes hashed_value: hashed value to compare with.

        :rtype: bool
        """

        return bcrypt.checkpw(text.encode(self._encoding), hashed_value)

    def _get_algorithm(self, **options):
        """
        gets the hashing algorithm.

        :rtype: str
        """

        return 'bcrypt'

    def _get_separator_count(self):
        """
        gets the separator count used between parts of this handler's hashed result.

        :rtype: int
        """

        return 4

    def _digest_inputs(self, text, **options):
        """
        validates given inputs for bcrypt hashing and
        returns the byte equivalent of text input.
        it raises an error on invalid inputs.

        :param str text: text to be hashed.

        :raises BcryptMaxSizeLimitError: bcrypt max size limit error.

        :rtype: bytes
        """

        text_bytes = text.encode(self._encoding)
        size = len(text_bytes)
        if size > self.MAX_SIZE:
            raise BcryptMaxSizeLimitError('Size of input string is [{input_size}] bytes '
                                          'which is larger than bcrypt\'s [{limit}] bytes limit.'
                                          .format(input_size=size,
                                                  limit=self.MAX_SIZE))

        return text_bytes

    def _get_hashed_part(self, full_hashed_value, **options):
        """
        gets the hashed part from full hashed value which current handler understands it.
        this method returns the original hash value made by bcrypt
        excluding the handler name.

        :param bytes full_hashed_value: full hashed value to get hashed part from it.

        :rtype: bytes
        """

        empty, handler, bcrypt_hash = full_hashed_value.split(self._get_separator(), 2)
        return self._get_separator() + bcrypt_hash

    def _make_final_hash(self, bcrypt_hash, **options):
        """
        makes final hash from input values and returns it.

        :param bytes bcrypt_hash: hash value which is generated by bcrypt algorithm.

        :rtype: bytes
        """

        return self._get_separator() + self._get_algorithm().encode(
            self._encoding) + bcrypt_hash



# -*- coding: utf-8 -*-
"""
security enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class TokenTypeEnum(CoreEnum):
    """
    token type enum.
    """

    ACCESS = 'access'
    REFRESH = 'refresh'



# -*- coding: utf-8 -*-
"""
sentry integrations sqlalchemy module.
"""

from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

from pyrin.logging.sentry.decorators import sentry_integration
from pyrin.logging.sentry.integrations.base import SentryIntegrationBase


@sentry_integration()
class SQLAlchemySentryIntegration(SentryIntegrationBase):
    """
    sqlalchemy sentry integration class.
    """

    name = 'sqlalchemy'

    def _configure(self, **options):
        """
        configures this integration and gets the configured object.

        :rtype: SqlalchemyIntegration
        """

        return SqlalchemyIntegration()



# -*- coding: utf-8 -*-
"""
encryption handlers base module.
"""

import re

from abc import abstractmethod

import pyrin.security.utils.services as security_utils_services
import pyrin.utils.encoding as encoding_utils

from pyrin.settings.static import APPLICATION_ENCODING
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.security.encryption.exceptions import DecryptionError
from pyrin.security.encryption.interface import AbstractEncrypterBase
from pyrin.security.encryption.handlers.exceptions import InvalidEncryptedValueError, \
    EncryptionHandlerMismatchError


class EncrypterBase(AbstractEncrypterBase):
    """
    encrypter base class.
    all application encrypters must subclass this.
    """

    # regular expression to validate format of full encrypted values.
    # the following format will be matched: `$handler_name$encrypted_value`
    FORMAT_REGEX = re.compile(r'^\$[^$]+\$(.+)$')

    def __init__(self, name, **options):
        """
        initializes an instance of EncrypterBase.

        :param str name: name of the encryption handler.
        """

        super().__init__()

        self._set_name(name)
        self._encoding = APPLICATION_ENCODING

    def _get_separator(self):
        """
        gets the separator used between each part of this handler's encryption result.

        :rtype: bytes
        """

        return b'$'

    def _get_separator_count(self):
        """
        gets the separator count used between parts of this handler's encryption result.

        :rtype: int
        """

        return 2

    @abstractmethod
    def _get_encryption_key(self, **options):
        """
        gets the signing key for encryption.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_decryption_key(self, **options):
        """
        gets the signing key for decryption.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    @abstractmethod
    def _get_algorithm(self, **options):
        """
        gets the algorithm used for encryption.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def encrypt(self, text, **options):
        """
        encrypts the given value and returns the full encrypted
        result which includes the handler name.

        :param str text: text to be encrypted.

        :rtype: str
        """

        encrypted = self._encrypt(text, **options)
        final_result = self._make_final_result(encrypted, **options)
        return self._prepare_output(final_result)

    @abstractmethod
    def _encrypt(self, text, **options):
        """
        encrypts the given value and returns the encrypted result.

        :param str text: text to be encrypted.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bytes
        """

        raise CoreNotImplementedError()

    def decrypt(self, full_encrypted_value, **options):
        """
        decrypts the given full encrypted value and returns the decrypted result.

        :param str full_encrypted_value: full encrypted value to be decrypted.

        :raises DecryptionError: decryption error.

        :rtype: str
        """

        try:
            self._validate_format(full_encrypted_value, **options)
            encrypted_part = self._get_encrypted_part(self._prepare_input(full_encrypted_value),
                                                      **options)

            return self._decrypt(encrypted_part, **options)
        except Exception as error:
            raise DecryptionError(error) from error

    @abstractmethod
    def _decrypt(self, value, **options):
        """
        decrypts the given value and returns the decrypted result.

        :param bytes value: value to be decrypted.

        :rtype: str
        """

        raise CoreNotImplementedError()

    def _get_encrypted_part(self, full_encrypted_value, **options):
        """
        gets the encrypted part from full encrypted value.

        :param bytes full_encrypted_value: full encrypted value to get encrypted part from it.

        :raises InvalidEncryptedValueError: invalid encrypted value error.
        :raises EncryptionHandlerMismatchError: encryption handler mismatch error.

        :rtype: bytes
        """

        handler, encrypted_part = self._extract_parts_from_final_value(
            full_encrypted_value, **options)

        self._validate_handler(handler, **options)

        return encrypted_part

    def _validate_format(self, full_encrypted_value, **options):
        """
        validates the format of full encrypted value.
        an exception will be raised on invalid values.

        :param str full_encrypted_value: full encrypted value to be validated.

        :raises InvalidEncryptedValueError: invalid encrypted value error.
        """

        if not self.FORMAT_REGEX.match(full_encrypted_value):
            raise InvalidEncryptedValueError('Input value is not a valid '
                                             '[{current}] encryption value.'
                                             .format(current=self._get_algorithm()))

    def _extract_parts_from_final_value(self, full_encrypted_value, **options):
        """
        extracts different parts from full encrypted value.

        :param bytes full_encrypted_value: full encrypted value to get different parts from it.

        :returns: tuple[str handler_name, bytes encrypted_part]
        :rtype: tuple[str, bytes]
        """

        empty, handler, encrypted_part = full_encrypted_value.split(
            self._get_separator(), self._get_separator_count())

        return handler.decode(self._encoding), self._decode_encrypted_part(encrypted_part)

    def _validate_handler(self, handler_name, **options):
        """
        validates the handler name to match the current handler.
        it will raise an error on handler mismatch.

        :param str handler_name: handler name to be validated.

        :raises EncryptionHandlerMismatchError: encryption handler mismatch error.
        """

        if handler_name != self._get_algorithm():
            raise EncryptionHandlerMismatchError('Encryption handler [{handler}] does not '
                                                 'match the current handler which is [{current}].'
                                                 .format(handler=handler_name,
                                                         current=self._get_algorithm()))

    def _make_final_result(self, encrypted_value, **options):
        """
        gets the final result by prepending the handler name to the encrypted value.

        :param bytes encrypted_value: encrypted value which handler name should be prepended to.

        :rtype: bytes
        """

        return self._get_separator() + self._get_separator().join(
            (self._get_algorithm().encode(self._encoding),
             self._encode_encrypted_part(encrypted_value)))

    def _prepare_output(self, full_encrypted_value, **options):
        """
        prepares output value before returning from this handler.

        :param bytes full_encrypted_value: full encrypted value to be prepared.

        :rtype: str
        """

        return full_encrypted_value.decode(self._encoding)

    def _prepare_input(self, full_encrypted_value, **options):
        """
        prepares input value to be handled by this handler.

        :param str full_encrypted_value: full encrypted value to be prepared.

        :rtype: bytes
        """

        return full_encrypted_value.encode(self._encoding)

    def _decode_encrypted_part(self, value):
        """
        decodes base64 encoded encrypted part into raw bytes.

        :param bytes value: base64 encoded value.

        :returns: raw bytes.
        :rtype: bytes
        """

        return encoding_utils.base64_to_bytes(value)

    def _encode_encrypted_part(self, value):
        """
        encodes raw bytes encrypted part into base64 encoded bytes.

        :param bytes value: raw bytes value.

        :returns: base64 encoded bytes.
        :rtype: bytes
        """

        return encoding_utils.bytes_to_base64(value)


class SymmetricEncrypterBase(EncrypterBase):
    """
    symmetric encrypter base class.
    this encrypter type uses a single symmetric key for encryption and decryption.
    """

    def __init__(self, **options):
        """
        initializes an instance of SymmetricEncrypterBase.
        """

        # we pass the algorithm of encryption handler as the name of it.
        super().__init__(self._get_algorithm(**options), **options)

    def _get_decryption_key(self, **options):
        """
        gets the signing key for decryption.

        :rtype: str
        """

        return self._get_encryption_key(**options)


class AsymmetricEncrypterBase(EncrypterBase):
    """
    asymmetric encrypter base class.
    this encrypter type uses a pair of public/private asymmetric
    keys for encryption and decryption.
    """

    def __init__(self, **options):
        """
        initializes an instance of AsymmetricEncrypterBase.
        """

        # we pass the algorithm of encryption handler as the name of it.
        super().__init__(self._get_algorithm(**options), **options)


class RSAEncrypterBase(AsymmetricEncrypterBase):
    """
    rsa encrypter base class.
    this encrypter type uses a pair of public/private asymmetric
    keys for encryption and decryption.
    """

    def __init__(self, **options):
        """
        initializes an instance of RSAEncrypterBase.
        """

        super().__init__(**options)

        self._private_key = None
        self._public_key = None

        self._load_keys(**options)

    def _get_encryption_key(self, **options):
        """
        gets the signing key for encryption.

        :rtype: object
        """

        return self._public_key

    def _get_decryption_key(self, **options):
        """
        gets the signing key for decryption.

        :rtype: object
        """

        return self._private_key

    def generate_key(self, **options):
        """
        generates a valid public/private key for this handler and returns it.

        :keyword int length: the length of generated key in bits.
                             if not provided, it uses default value
                             from relevant config.

        :returns: tuple[str public_key, str private_key]
        :rtype: tuple[str, str]
        """

        return security_utils_services.generate_rsa_key(**options)

    @abstractmethod
    def _load_keys(self, **options):
        """
        loads public/private keys into this class's relevant attributes.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
celery cli package.
"""

from pyrin.packaging.base import Package


class CeleryCLIPackage(Package):
    """
    celery cli package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'task_queues.celery.cli.component'
    DEPENDS = ['pyrin.cli']



# -*- coding: utf-8 -*-
"""
orm sql extractor package.
"""

from pyrin.packaging.base import Package


class ORMSQLExtractorPackage(Package):
    """
    orm sql extractor package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.orm.sql.extractor.component'
